#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from dataclasses import dataclass, field
from typing import Self, Callable, Any, TypeVar

from frozendict import frozendict
from init_helpers import Jsonable, raise_if, AnyType

from openapi_tools.spec import SpecRef
from openapi_tools.spec.info import Info
from openapi_tools.spec.maybe_referencable_resource import MaybeReferencableResource
from openapi_tools.spec.spec_resource import SpecResource

K = TypeVar('K')
V = TypeVar('V')


class MappedDict(dict[K, V]):
    def __init__(self, mapper: Callable[[Any], K], extract_controller: Callable[[K], bool]) -> None:
        super().__init__()
        self.mapper = mapper
        self.extract_controller = extract_controller

    def __getitem__(self, key: Any) -> V:
        mapped_key = self.mapper(key)
        if self.extract_controller(mapped_key):
            return dict.__getitem__(self, mapped_key)
        return mapped_key

    def __setitem__(self, key: Any, value) -> None:
        dict.__setitem__(self, self.mapper(key), value)


@dataclass
class Spec:
    schema_builder: Callable[[AnyType], SpecResource]
    payload: dict[str, Jsonable] = field(init=False)
    resource_to_ref: dict[SpecResource, SpecRef | frozendict] = field(init=False)

    def __init__(self, info: Info, schema_builder: Callable[[AnyType], SpecResource]):
        self.schema_builder = schema_builder
        self.payload = {'openapi': '3.1.0', 'info': info.get_spec_dict({}), 'paths': {}}
        self.resource_to_ref = MappedDict(mapper=schema_builder,
                                          extract_controller=lambda x: isinstance(x, SpecResource))

    def add_resource(self: Self, resource: MaybeReferencableResource, key: str | None = None) -> None:
        key: str = key or resource.get_key()
        raise_if(key is None, ValueError(f'{key=} is not defined'))
        ref = resource.get_spec_ref()
        target: dict[str, Jsonable] = self.payload
        for step in resource.spec_path:
            target = target.setdefault(step, {})

        target[key] = (
            resource.merge_spec_dict(self.resource_to_ref, present_value)
            if (present_value := target.get(key))
            else resource.get_spec_dict(self.resource_to_ref))
        self.resource_to_ref[resource] = ref

    def prepare_resource(self: Self, resource: SpecResource | AnyType) -> None:
        if resource in self.resource_to_ref:
            return
        resource = self.schema_builder(resource)
        self.resource_to_ref[resource] = resource.get_spec_dict(self.resource_to_ref)

    def has_ref(self: Self, ref: SpecRef) -> bool:
        spec_path = ref.ref.lstrip("#/").split('/')
        assert spec_path, f'bad {spec_path=}'
        target: dict[str, Jsonable] = self.payload
        for step in spec_path[:-1]:
            target = target.setdefault(step, {})
        return spec_path[-1] in target



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import functools
from dataclasses import dataclass
from typing import Iterable, Mapping

from frozendict import frozendict
from init_helpers import AnyType, Jsonable
from more_itertools import first

from .example import BaseExample
from .schema import BaseSchema
from .spec_ref import SpecRef
from .spec_resource import SpecResource


@dataclass(frozen=True, slots=True)
class MediaType(SpecResource):
    schema: AnyType | BaseSchema
    examples: Iterable[BaseExample] = tuple()
    # encoding: Mapping[str, EncodingObject] = field(default_factory=frozendict) todo: implement encoding

    def __post_init__(self):
        object.__setattr__(self, 'examples', tuple(self.examples))

    def get_spec_dependencies(self) -> frozenset['SpecResource']:
        return frozenset({self.schema} | set(self.examples))

    def get_spec_dict(self, dependency_to_ref: Mapping['SpecResource', SpecRef]) -> frozendict[str, Jsonable]:
        result = {'schema': dependency_to_ref[self.schema]}
        if self.examples:
            result['example'] = dependency_to_ref[first(self.examples)]
        return frozendict(result)

    @functools.cache
    def _get_repr_parts(self) -> tuple[str, ...]:
        result = [f'{self.schema!r}']
        result += [f'examples={self.examples!r}'] if self.examples else []
        return tuple(result)

    def __repr__(self):
        return f'{self.__class__.__name__}({", ".join(self._get_repr_parts())})'

    __str__ = __repr__



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from dataclasses import dataclass

from init_helpers import ReprInDumps


@dataclass(frozen=True, slots=True)
class SpecRef(ReprInDumps):
    ref: str
    summary: str = ''
    description: str = ''

    def __repr_in_dumps__(self) -> dict[str, str]:
        result = {'$ref': self.ref}
        result |= {'summary': self.summary} if self.summary else {}
        result |= {'description': self.description} if self.description else {}
        return result

    def __repr__(self) -> str:
        return f'SpecRef({", ".join({f"{key}={value!r}" for key, value in self.__repr_in_dumps__().items()})})'

    __str__ = __repr__



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import asyncio
import datetime
import logging
from dataclasses import dataclass, field
from typing import Callable, Awaitable, Iterable, ClassVar, Generic, Mapping, Any

import aiohttp
from async_tools import Periodic
from authlib.jose import JsonWebToken, JWTClaims, JoseError, JsonWebKey, KeySet
from frozendict import frozendict
from http_tools import IncomingRequest
from init_helpers import raise_if, Jsonable, url_to_snake_case

from openapi_tools.spec import SpecResource, SpecRef
from ..exceptions import Unauthorized
from .security_scheme import SecurityScheme, AuthInfo, AuthToken
from .security_scheme_type import SecuritySchemeType

logger = logging.getLogger(__name__)


@dataclass(frozen=True, kw_only=True)
class OpenIdConnectSecurityScheme(SecurityScheme, Generic[AuthInfo]):
    type_: ClassVar[SecuritySchemeType] = SecuritySchemeType.open_id_connect
    header_scheme: ClassVar[str] = 'Bearer'
    jwks_refresh_period: ClassVar[datetime.timedelta] = datetime.timedelta(seconds=60)
    jwt: ClassVar[JsonWebToken] = JsonWebToken(['RS256'])

    url: str
    session: aiohttp.ClientSession | None = None
    jwks_request_kwargs: frozendict[str, Any] = field(default_factory=lambda: frozendict(timeout=10))
    resolver: Callable[[AuthToken], Awaitable[tuple[AuthInfo, Iterable[str]]] | tuple[AuthInfo, Iterable[str]]] = (
        field(init=False, repr=False, compare=False))
    expected_issuer: str = field(init=False, repr=False, compare=False)
    periodic_refresh_public_keys: Periodic = field(init=False, repr=False, compare=False)
    jwks: KeySet = field(init=False, repr=False, compare=False)
    jwks_loaded_event: asyncio.Event = field(init=False, default_factory=asyncio.Event, repr=False, compare=False)

    def __post_init__(self):
        well_known_index = self.url.find('/.well-known')
        raise_if(well_known_index == -1, ValueError('OpenIdConnectSecurityScheme URL requires ".well-known" part'))
        expected_issuer = self.url[:well_known_index]
        object.__setattr__(self, 'expected_issuer', expected_issuer)
        object.__setattr__(self, 'periodic_refresh_public_keys', Periodic(
            self._refresh_public_keys, self.jwks_refresh_period, first_at=datetime.datetime.now()))

        if self.session is None:
            logger.warning('OpenIdConnectSecurityScheme cannot await session.close(), '
                           'so do it manually or pass the session controlled from outer scope')
            object.__setattr__(self, 'session', aiohttp.ClientSession())

        async def resolver(jwt_value: str) -> tuple[AuthInfo, frozenset[str]]:
            try:
                claims: JWTClaims = self.jwt.decode(jwt_value, await self.get_jwks())
                self._check_claims(claims)
            except (JoseError, ValueError) as e:
                raise Unauthorized(str(e)) from None

            return claims, self._get_scopes(claims)

        object.__setattr__(self, 'resolver', resolver)

    @staticmethod
    def _get_scopes(claims: JWTClaims) -> frozenset[str]:
        if (scope := claims.get('scope')) is None:
            raise Unauthorized(f'BadJWT: Missing scope')
        if not isinstance(scope, str):
            raise Unauthorized(f'BadJWT: Scope is not string: {type(scope).__name__}')
        scopes = frozenset(scope.split(' '))

        return scopes

    def _check_claims(self, claims: JWTClaims) -> None:
        claims.validate()
        if (issuer := claims.get('iss')) != self.expected_issuer:
            raise Unauthorized(f'BadJWT: Invalid {issuer=}, expected: {self.expected_issuer!r}')
        # TODO: add "aud" check
        # if 'your-client-id' not in (audience := claims.get('aud', [])):
        #     raise Unauthorized(f'BadJWT: Invalid {audience=}')

    async def _extract_token(self, incoming_request: IncomingRequest) -> AuthToken:
        dirty_token: str = incoming_request.metadata.header_name_to_value.get("Authorization")
        raise_if(not isinstance(dirty_token, str), Unauthorized("Failed to get token from Authorization header"))
        raise_if(not dirty_token.startswith(self.header_scheme), Unauthorized("Wrong authorization scheme"))
        return dirty_token.removeprefix(self.header_scheme).strip()

    def get_key(self) -> str:
        return f'openid__{url_to_snake_case(self.expected_issuer)}'

    def get_spec_dict(self, dependency_to_ref: Mapping[SpecResource, SpecRef]) -> frozendict[str, Jsonable]:
        return SecurityScheme.get_spec_dict(self, dependency_to_ref=dependency_to_ref) | {'openIdConnectUrl': self.url}

    async def get_public_keys_url(self) -> str:
        if not hasattr(self, 'public_keys_url'):
            logger.debug(f'get_public_keys_url from {self.url}')
            try:
                async with self.session.get(self.url, **self.jwks_request_kwargs) as resp:
                    configuration = await resp.json()
                object.__setattr__(self, 'public_keys_url', configuration['jwks_uri'])
            except Exception as e:
                raise RuntimeError(f"Failed to get public keys url: {e!r}") from e

        return getattr(self, 'public_keys_url')

    async def _load_public_keys(self) -> list[dict]:
        logger.debug(f'get_public_keys_url from {(public_keys_url := await self.get_public_keys_url())}')
        async with self.session.get(public_keys_url, **self.jwks_request_kwargs) as resp:
            jwks = await resp.json()
        return jwks['keys']

    async def _refresh_public_keys(self) -> None:
        logger.debug('_refresh_public_keys')
        object.__setattr__(self, 'jwks', JsonWebKey.import_key_set({'keys': await self._load_public_keys()}))
        self.jwks_loaded_event.set()

    async def get_jwks(self) -> KeySet:
        await self.jwks_loaded_event.wait()
        return self.jwks



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from .requirement import SecurityRequirement
from .scheme import ApiKeySecurityScheme
from .scheme import HttpSecurityScheme, HttpBearerSecurityScheme, HttpBasicSecurityScheme
from .scheme import OAuth2Flow
from .scheme import OAuth2SecurityScheme
from .scheme import OpenIdConnectSecurityScheme
from .scheme import SecurityScheme, AuthInfo, AuthToken, resolver_cache
from .scheme import SecuritySchemeType
from .exceptions import SecurityException, Unauthorized, Forbidden, AuthInfoException
from .extras import get_securities_arg_names



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import enum


class SchemaFormat(enum.StrEnum):
    pass


class IntegerSchemaFormat(SchemaFormat):
    int32 = 'int32'  # signed 32 bits integer
    int64 = 'int64'  # signed 64 bits integer


class NumberSchemaFormat(SchemaFormat):
    float = 'float'  # signed 32 bits float
    double = 'double'  # signed 64 bits float


class StringSchemaFormat(SchemaFormat):
    date = 'date'  # YYYY-MM-DD: "1970-01-01" / "2025-07-01"
    time = 'time'  # hh:mm:ss.sTZD: "10:05:08" / "10:05:08.5" / "10:05:08-02:30" / "10:05:08Z"
    date_time = 'date-time'  # YYYY-MM-DDThh:mm:ss.sTZD:
    # "1970-01-01T10:05:08" / "2025-07-01T10:05:08.10" / "2025-07-01T10:05:08+01:00"
    duration = 'duration'  # ISO 8601: "P4DT12H30M5S"
    regex = 'regex'  # valid re: "^[a-z]+$"
    email = 'email'  # ascii email: "john@example.com"
    idn_email = 'idn-email'  # Internationalized Domain Name email: "иван@иванов.ру"
    hostname = 'hostname'  # ascii hostname(could be encoded): "www.example.com" / "xn--4gbwdl.xn--wgbh1c"
    idn_hostname = 'idn-hostname'  # Internationalized Domain Name hostname: "иванов.ру"
    ipv4 = 'ipv4'  # ipv4: "192.168.0.1"
    ipv6 = 'ipv6'  # ipv6: "::1" / "2001:db8:3333:4444:5555:6666:7777:8888"
    json_pointer = 'json-pointer'  # absolute json pointer(starting with '/'): "/a/b/c"
    relative_json_pointer = 'relative-json-pointer'  # relative json pointer: "0/a/b" / "5/a/b#" / "2#"
    uri = 'uri'  # full ascii uri: "http://example.com/path?qs=v&qs2[1]=3#fragment" / "http://example.com/index.html"
    uri_reference = 'uri-reference'  # relative ascii uri: "aaa/bbb.html" / "?a=42" / "http://example.com/index.html"
    uri_template = 'uri-template'  # uri template: "/{+file}.html" / "{?q,lang}"
    iri = 'iri'  # Internationalized Resource Identifier: "http://иванов.ру"
    iri_reference = 'iri-reference'  # iri ref: "/иванов.ру" / "#фрагмент"
    uuid = 'uuid'  # any ver/case:  "2EB8AA08-AA98-11EA-B4AA-73B441D16380" / "2eb8aa08-aa98-11ea-b4aa-73b441d16380"
    # OpenAPI extension to JsonSchema formats
    password = 'password'  # A hint to UIs to obscure input.
    binary = 'binary'  # format for bytes



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import functools
from abc import ABC
from dataclasses import dataclass
from typing import Mapping, ClassVar

from frozendict import frozendict
from init_helpers import Jsonable

from openapi_tools.spec.spec_resource import SpecResource, SpecRef
from openapi_tools.spec.schema.base_schema import BaseSchema
from openapi_tools.spec.schema.schema_format import SchemaFormat


@dataclass(frozen=True, slots=True)
class TypeSchema(BaseSchema, ABC):
    type_: ClassVar[str]
    format_: SchemaFormat | None = None

    def get_spec_dict(self, dependency_to_ref: Mapping['SpecResource', SpecRef]) -> frozendict[str, Jsonable]:
        result: dict[str, Jsonable] = {'type': self.type_}
        if self.format_ is not None:
            result['format'] = self.format_
        return frozendict(result | BaseSchema.get_spec_dict(self, dependency_to_ref))

    @functools.cache
    def _get_repr_parts(self) -> tuple[str, ...]:
        parts = list(BaseSchema._get_repr_parts(self))
        parts += [f'format_={self.format_!r}'] if self.format_ else []
        return tuple(parts)



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from enum import StrEnum


class ParameterLocation(StrEnum):
    query = 'query'
    path = 'path'
    header = 'header'
    cookie = 'cookie'
    body = 'body'  # INNER: not present in specification, used inside library
    security = 'security'  # INNER: not present in specification, used inside library
    literal = 'literal'  # INNER: not present in specification, used inside library



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from .aggregation import CallTemplate, ParameterAggregation
from .body import BodyParameter, RawBodyParameter, JsonBodyParameter, JsonWholeBodyParameter
from .call_parameter import CallParameter
from .cookie_parameter import CookieParameter
from .header_parameter import HeaderParameter
from .literal_parameter import LiteralParameter
from .multipart_parameter import MultipartParameter
from .parameter_location import ParameterLocation
from .parameter_style import ParameterStyle
from .path_parameter import PathParameter
from .query_parameter import QueryParameter
from .security_parameter import SecurityParameter
from .spec_parameter import SpecParameter



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import dataclasses
import functools
from dataclasses import dataclass, MISSING
from typing import ClassVar, Mapping

from frozendict import frozendict
from init_helpers import Jsonable

from openapi_tools.spec.spec_resource import SpecResource, SpecRef
from .schema_format import IntegerSchemaFormat
from .type_schema import TypeSchema

@dataclass(frozen=True, slots=True)
class IntegerSchema(TypeSchema):
    type_: ClassVar[str] = "integer"
    format_: IntegerSchemaFormat | None = None
    exclusive_minimum: int | None = None
    exclusive_maximum: int | None = None

    def get_spec_dict(self, dependency_to_ref: Mapping['SpecResource', SpecRef | dict]) -> frozendict[str, Jsonable]:
        result = {}
        if self.exclusive_minimum is not None:
            result |= {'exclusiveMinimum': self.exclusive_minimum}
        if self.exclusive_maximum is not None:
            result |= {'exclusiveMaximum': self.exclusive_maximum}
        return frozendict(TypeSchema.get_spec_dict(self, dependency_to_ref) | result)

    @functools.cache
    def _get_repr_parts(self) -> tuple[str, ...]:
        parts = list(TypeSchema._get_repr_parts(self))
        parts += [f'exclusive_minimum={self.exclusive_minimum!r}'] if self.exclusive_minimum else []
        parts += [f'exclusive_maximum={self.exclusive_maximum!r}'] if self.exclusive_maximum else []
        return tuple(parts)

    def __repr__(self):
        return f'{self.__class__.__name__}({", ".join(self._get_repr_parts())})'

    __str__ = __repr__

#
# print('ih')
#
#
# def _tuple_str(obj_name, fields):
#     # Return a string representing each field of obj_name as a tuple
#     # member.  So, if fields is ['x', 'y'] and obj_name is "self",
#     # return "(self.x,self.y)".
#
#     # Special case for the 0-tuple.
#     if not fields:
#         return '()'
#     # Note the trailing comma, needed if this turns out to be a 1-tuple.
#     return f'({",".join([f"{obj_name}.{f.name}" for f in fields])},)'
#
# # field_list = [f for f in fields.values() if f._field_type is _FIELD]
#
#
# def _create_fn(name, args, body, *, globals=None, locals=None,
#                return_type=MISSING):
#     # Note that we may mutate locals. Callers beware!
#     # The only callers are internal to this module, so no
#     # worries about external callers.
#     if locals is None:
#         locals = {}
#     return_annotation = ''
#     if return_type is not MISSING:
#         locals['_return_type'] = return_type
#         return_annotation = '->_return_type'
#     args = ','.join(args)
#     body = '\n'.join(f'  {b}' for b in body)
#
#     # Compute the text of the entire function.
#     txt = f' def {name}({args}){return_annotation}:\n{body}'
#
#     local_vars = ', '.join(locals.keys())
#     txt = f"def __create_fn__({local_vars}):\n{txt}\n return {name}"
#     ns = {}
#     exec(txt, globals, ns)
#     return ns['__create_fn__'](**locals)
#
#
# def __create_fn__():
#  def __eq__(self,other):
#   if other.__class__ is self.__class__:
#    return (self.default,self.description,self.examples,self.key,self.enum,self.format_,self.exclusive_minimum,self.exclusive_maximum,)==(other.default,other.description,other.examples,other.key,other.enum,other.format_,other.exclusive_minimum,other.exclusive_maximum,)
#   return NotImplemented
#  return __eq__
#
# def _cmp_fn(name, op, self_tuple, other_tuple, globals):
#     # Create a comparison function.  If the fields in the object are
#     # named 'x' and 'y', then self_tuple is the string
#     # '(self.x,self.y)' and other_tuple is the string
#     # '(other.x,other.y)'.
#
#     return _create_fn(name,
#                       ('self', 'other'),
#                       [ 'if other.__class__ is self.__class__:',
#                        f' return {self_tuple}{op}{other_tuple}',
#                         'return NotImplemented'],
#                       globals=globals)
#
# field_list = dataclasses.fields(IntegerSchema)
#
# # Create __eq__ method.  There's no need for a __ne__ method,
# # since python will call __eq__ and negate it.
# flds = [f for f in field_list if f.compare]
# self_tuple = _tuple_str('self', flds)
# other_tuple = _tuple_str('other', flds)
# _cmp_fn('__eq__', '==', self_tuple, other_tuple, globals=globals)
#
# # IntegerSchema = dataclasses._process_class(
# #     IntegerSchema,
# #     frozen=True,
# #     slots=True,
# #
# #     init=True,
# #     repr=True,
# #     eq=True,
# #     order=False,
# #     unsafe_hash=False,
# #     # frozen=False,
# #     match_args=True,
# #     kw_only=False,
# #     # slots=False,
# #     weakref_slot=False
# # )


import inspect
import string
from inspect import Signature
from typing import Callable, Literal, Annotated, Iterable

from http_tools import Answer, WrappedAnswer, HttpStatusCode
from http_tools.answer import ExceptionAnswer, JsonableAnswer
from init_helpers import raise_if, try_extract_type_notes

from ..spec import Tag
from ..spec import Example, SecurityRequirement, get_securities_arg_names, SecurityScheme
from ..spec import SpecParameter, QueryParameter, PathParameter, JsonBodyParameter, CallParameter, RawBodyParameter
from .endpoint import AnswerType, ResultType, Endpoint


def gen_query_endpoint(
        func: Callable,
        method: str, path: str = '', securities: Iterable[SecurityRequirement | SecurityScheme] = tuple(),
        exception_type_to_answer_type: dict[type[Exception], type[Answer]] | None = None,
        unhandled_exception__answer_type: type[Answer] | None = ExceptionAnswer[HttpStatusCode.InternalServerError],
        arg_name_to_param: dict[str, CallParameter | type[RawBodyParameter] | None] | None = None,
        answer_type: type[AnswerType] = WrappedAnswer[HttpStatusCode.OK],
        answer_factory: Callable[[ResultType], AnswerType] | None = None,
        summary: str | None = None,
        description: str | None = None,
        tags: list[str | Tag] | None = None,
        deprecated: bool = False,
        operation_id: str = '',
        ignore_extra_params: bool = False,
        additional_parameters: tuple[SpecParameter, ...] | None = None,  # add to spec, no other processing
) -> Endpoint:
    """
    Generate Endpoint based on function signature:
    - Return type will be used to generated answer type.
    - All missing arguments will be treated as QueryParameters
    """
    return _gen_endpoint(
        param_type=QueryParameter, func=func, method=method, path=path, securities=securities,
        exception_type_to_answer_type=exception_type_to_answer_type,
        unhandled_exception__answer_type=unhandled_exception__answer_type,
        arg_name_to_param=arg_name_to_param, answer_type=answer_type, answer_factory=answer_factory,
        summary=summary, description=description, tags=tags, deprecated=deprecated, operation_id=operation_id,
        ignore_extra_params=ignore_extra_params, additional_parameters=additional_parameters,
    )


def gen_json_endpoint(
        func: Callable,
        method: str, path: str = '', securities: Iterable[SecurityRequirement | SecurityScheme] = tuple(),
        exception_type_to_answer_type: dict[type[Exception], type[Answer]] | None = None,
        unhandled_exception__answer_type: type[Answer] | None = ExceptionAnswer[HttpStatusCode.InternalServerError],
        arg_name_to_param: dict[str, CallParameter | type[RawBodyParameter] | None] | None = None,
        answer_type: type[AnswerType] = WrappedAnswer[HttpStatusCode.OK],
        answer_factory: Callable[[ResultType], AnswerType] | None = None,
        summary: str | None = None,
        description: str | None = None,
        tags: list[str | Tag] | None = None,
        deprecated: bool = False,
        operation_id: str = '',
        ignore_extra_params: bool = False,
        additional_parameters: tuple[SpecParameter, ...] | None = None,  # add to spec, no other processing
) -> Endpoint:
    """
    Generate RPC endpoint based on function signature:
    - Return type will be used to generated answer type.
    - All missing arguments will be treated as JsonBodyParameter
    """
    return _gen_endpoint(
        param_type=JsonBodyParameter, func=func, method=method, path=path, securities=securities,
        exception_type_to_answer_type=exception_type_to_answer_type,
        unhandled_exception__answer_type=unhandled_exception__answer_type,
        arg_name_to_param=arg_name_to_param, answer_type=answer_type, answer_factory=answer_factory,
        summary=summary, description=description, tags=tags, deprecated=deprecated, operation_id=operation_id,
        ignore_extra_params=ignore_extra_params, additional_parameters=additional_parameters,
    )


def _get_path_param_names(path: str) -> set[str]:
    return {i[1] for i in string.Formatter().parse(path) if i[1] is not None}


def _gen_endpoint(
        param_type: type[SpecParameter], func: Callable[..., ResultType],
        method: str, path: str = '', securities: Iterable[SecurityRequirement | SecurityScheme] = tuple(),
        exception_type_to_answer_type: dict[type[Exception], type[Answer]] | None = None,
        unhandled_exception__answer_type: type[Answer] | None = ExceptionAnswer[HttpStatusCode.InternalServerError],
        arg_name_to_param: dict[str, CallParameter | type[RawBodyParameter]] | None = None,
        answer_type: type[AnswerType] = WrappedAnswer[HttpStatusCode.OK],
        answer_factory: Callable[[ResultType], AnswerType] | None = None,
        summary: str | None = None,
        description: str | None = None,
        tags: list[str | Tag] | None = None,
        deprecated: bool = False,
        operation_id: str = '',
        ignore_extra_params: bool = False,
        additional_parameters: tuple[SpecParameter, ...] | None = None,  # add to spec, no other processing
) -> Endpoint:
    path_param_names = _get_path_param_names(path)
    security_param_names = get_securities_arg_names(securities)
    signature = Signature.from_callable(func)
    arg_name_to_param = arg_name_to_param or {}
    result_arg_name_to_param = {}
    for arg in signature.parameters.values():
        if arg.annotation == Signature.empty:
            raise TypeError(f"undefined type for arg: {arg}")

        # дописать description и примеры?
        if arg.name in arg_name_to_param:
            if arg_name_to_param[arg.name] is not None:
                result_arg_name_to_param[arg.name] = arg_name_to_param[arg.name]
                # if
            continue

        schema = arg.annotation
        schema, notes = try_extract_type_notes(schema)
        if arg.kind == inspect.Parameter.VAR_POSITIONAL:
            schema = list[schema]
        elif arg.kind == inspect.Parameter.VAR_KEYWORD:
            schema = dict[str, schema]

        parameter_kwargs = {"name": arg.name, "schema": schema}
        if examples := tuple(note for note in notes if isinstance(note, Example)):
            parameter_kwargs["examples"] = examples

        if descriptions := [note for note in notes if isinstance(note, str)]:
            parameter_kwargs["description"] = descriptions[0]

        # description: str = ''
        # examples: tuple[BaseExample, ...] = field(default_factory=tuple)
        if arg.default != Signature.empty:
            parameter_kwargs["default"] = arg.default
        if arg.name in path_param_names:
            param = PathParameter(**parameter_kwargs)
        elif arg.name in security_param_names:
            continue
        else:
            # noinspection PyArgumentList
            param = param_type(**parameter_kwargs)
        result_arg_name_to_param[arg.name] = param

    if not ignore_extra_params:
        if extra_params := {k for k, v in arg_name_to_param.items() if v is not None} - result_arg_name_to_param.keys():
            raise KeyError(f"arg_name_to_param has {extra_params=}, use ignore_extra_params=True to skip those")

    if signature.return_annotation == Signature.empty:
        raise TypeError("undefined return type")

    answer_type, notes = try_extract_type_notes(answer_type)
    if issubclass(answer_type, JsonableAnswer) and not isinstance(answer_type.get_class_payload_type(), type):
        return_type = signature.return_annotation
        return_type = type[None] if return_type is None else return_type
        answer_type = answer_type[return_type]
    if notes:
        answer_type = Annotated[answer_type, *notes]

    raise_if(answer_type.get_class_status_code() is None,
             TypeError(f'Answer type MUST have status code, got {answer_type}'))

    # raise_if(answer_type.get_class_content_type() is None,
    #          TypeError(f'Answer type MUST have content type, got {answer_type}'))
    return Endpoint(
        func=func, method=method, path=path, securities=securities, arg_name_to_param=result_arg_name_to_param,
        answer_type=answer_type,
        exception_type_to_answer_type=exception_type_to_answer_type,
        unhandled_exception__answer_type=unhandled_exception__answer_type,
        answer_factory=answer_factory,
        summary=summary, description=description, tags=tags, deprecated=deprecated, operation_id=operation_id,
        additional_parameters=additional_parameters
    )
